"""Core module for testing the helper functions."""

import json
from pathlib import Path
from typing import Generator
from unittest import mock

import pytest

from vantage_agent.helpers import (
    cache_dict,
    load_cached_dict,
    parse_slurm_config,
    parse_slurm_nodes,
    parse_slurm_partitions,
    parse_slurm_queue,
)


@pytest.fixture
def raw_scontrol_show_config() -> Generator[str, None, None]:
    """Yield an example of the `scontrol show config` command."""
    yield """
    Configuration data as of 2024-08-19T19:08:02
    AccountingStorageBackupHost = (null)
    AccountingStorageEnforce = none
    AccountingStorageHost   = 10.59.51.10
    AccountingStorageExternalHost = (null)
    AccountingStorageParameters = (null)
    AccountingStoragePort   = 6839
    AccountingStorageTRES   = cpu,mem,energy,node,billing,fs/disk,vmem,pages
    AccountingStorageType   = accounting_storage/slurmdbd
    AccountingStorageUser   = N/A
    AccountingStoreFlags    = (null)
    AcctGatherEnergyType    = acct_gather_energy/none
    AcctGatherFilesystemType = acct_gather_filesystem/none
    AcctGatherInterconnectType = acct_gather_interconnect/none
    AcctGatherNodeFreq      = 0 sec
    AcctGatherProfileType   = acct_gather_profile/none
    AllowSpecResourcesUsage = No
    AuthAltTypes            = auth/jwt
    AuthAltParameters       = jwt_key=/var/lib/slurm/slurmctld/jwt_hs256.key
    AuthInfo                = socket=/var/run/munge/munge.socket.2
    AuthType                = auth/munge
    BatchStartTimeout       = 10 sec
    BcastExclude            = /lib,/usr/lib,/lib64,/usr/lib64
    BcastParameters         = (null)
    BOOT_TIME               = 2024-08-07T19:57:56
    BurstBufferType         = (null)
    CliFilterPlugins        = (null)
    ClusterName             = demo-cluster
    CommunicationParameters = (null)
    CompleteWait            = 0 sec
    CoreSpecPlugin          = core_spec/none
    CpuFreqDef              = Unknown
    CpuFreqGovernors        = OnDemand,Performance,UserSpace
    CredType                = cred/munge
    DebugFlags              = NO_CONF_HASH
    DefMemPerNode           = UNLIMITED
    DependencyParameters    = (null)
    DisableRootJobs         = No
    EioTimeout              = 60
    EnforcePartLimits       = NO
    Epilog                  = (null)
    EpilogMsgTime           = 2000 usec
    EpilogSlurmctld         = (null)
    ExtSensorsType          = ext_sensors/none
    ExtSensorsFreq          = 0 sec
    FederationParameters    = (null)
    FirstJobId              = 1
    GetEnvTimeout           = 2 sec
    GresTypes               = (null)
    GpuFreqDef              = (null)
    GroupUpdateForce        = 1
    GroupUpdateTime         = 600 sec
    HASH_VAL                = Match
    HealthCheckInterval     = 0 sec
    HealthCheckNodeState    = ANY
    HealthCheckProgram      = (null)
    InactiveLimit           = 0 sec
    InteractiveStepOptions  = --interactive --preserve-env --pty $SHELL
    JobAcctGatherFrequency  = 30
    JobAcctGatherType       = jobacct_gather/linux
    JobAcctGatherParams     = (null)
    JobCompHost             = localhost
    JobCompLoc              = (null)
    JobCompParams           = (null)
    JobCompPort             = 0
    JobCompType             = jobcomp/none
    JobCompUser             = root
    JobContainerType        = job_container/none
    JobCredentialPrivateKey = (null)
    JobCredentialPublicCertificate = (null)
    JobDefaults             = (null)
    JobFileAppend           = 0
    JobRequeue              = 1
    JobSubmitPlugins        = (null)
    KillOnBadExit           = 0
    KillWait                = 30 sec
    LaunchParameters        = (null)
    Licenses                = (null)
    LogTimeFormat           = iso8601_ms
    MailDomain              = (null)
    MailProg                = /bin/mail
    MaxArraySize            = 1001
    MaxBatchRequeue         = 5
    MaxDBDMsgs              = 20004
    MaxJobCount             = 10000
    MaxJobId                = 67043328
    MaxMemPerNode           = UNLIMITED
    MaxNodeCount            = 1
    MaxStepCount            = 40000
    MaxTasksPerNode         = 512
    MCSPlugin               = mcs/none
    MCSParameters           = (null)
    MessageTimeout          = 10 sec
    MinJobAge               = 300 sec
    MpiDefault              = none
    MpiParams               = (null)
    MULTIPLE_SLURMD         = Yes
    NEXT_JOB_ID             = 1
    NodeFeaturesPlugins     = (null)
    OverTimeLimit           = 0 min
    PluginDir               = /usr/lib/x86_64-linux-gnu/slurm-wlm/
    PlugStackConfig         = /etc/slurm/plugstack.conf
    PowerParameters         = (null)
    PowerPlugin             =
    PreemptMode             = OFF
    PreemptParameters       = (null)
    PreemptType             = preempt/none
    PreemptExemptTime       = 00:00:00
    PrEpParameters          = (null)
    PrEpPlugins             = prep/script
    PriorityParameters      = (null)
    PrioritySiteFactorParameters = (null)
    PrioritySiteFactorPlugin = (null)
    PriorityType            = priority/basic
    PrivateData             = none
    ProctrackType           = proctrack/linuxproc
    Prolog                  = (null)
    PrologEpilogTimeout     = 65534
    PrologSlurmctld         = (null)
    PrologFlags             = (null)
    PropagatePrioProcess    = 0
    PropagateResourceLimits = ALL
    PropagateResourceLimitsExcept = (null)
    RebootProgram           = (null)
    ReconfigFlags           = (null)
    RequeueExit             = (null)
    RequeueExitHold         = (null)
    ResumeFailProgram       = (null)
    ResumeProgram           = (null)
    ResumeRate              = 300 nodes/min
    ResumeTimeout           = 60 sec
    ResvEpilog              = (null)
    ResvOverRun             = 0 min
    ResvProlog              = (null)
    ReturnToService         = 2
    RoutePlugin             = route/default
    SchedulerParameters     = (null)
    SchedulerTimeSlice      = 30 sec
    SchedulerType           = sched/backfill
    ScronParameters         = (null)
    SelectType              = select/cons_res
    SelectTypeParameters    = CR_CORE
    SlurmUser               = slurm(64030)
    SlurmctldAddr           = 10.59.51.10
    SlurmctldDebug          = info
    SlurmctldHost[0]        = democluster-example
    SlurmctldLogFile        = /var/log/slurm/slurmctld.log
    SlurmctldPort           = 6817
    SlurmctldSyslogDebug    = (null)
    SlurmctldPrimaryOffProg = (null)
    SlurmctldPrimaryOnProg  = (null)
    SlurmctldTimeout        = 300 sec
    SlurmctldParameters     = (null)
    SlurmdDebug             = info
    SlurmdLogFile           = /var/log/slurm/slurmd.log
    SlurmdParameters        = (null)
    SlurmdPidFile           = /var/run/slurmd.pid
    SlurmdPort              = 6818
    SlurmdSpoolDir          = /var/lib/slurm/slurmd
    SlurmdSyslogDebug       = (null)
    SlurmdTimeout           = 60 sec
    SlurmdUser              = root(0)
    SlurmSchedLogFile       = (null)
    SlurmSchedLogLevel      = 0
    SlurmctldPidFile        = /var/run/slurmctld.pid
    SLURM_CONF              = /etc/slurm/slurm.conf
    SLURM_VERSION           = 23.02.7
    SrunEpilog              = (null)
    SrunPortRange           = 0-0
    SrunProlog              = (null)
    StateSaveLocation       = /var/lib/slurm/checkpoint
    SuspendExcNodes         = (null)
    SuspendExcParts         = (null)
    SuspendExcStates        = (null)
    SuspendProgram          = (null)
    SuspendRate             = 60 nodes/min
    SuspendTime             = INFINITE
    SuspendTimeout          = 30 sec
    SwitchParameters        = (null)
    SwitchType              = switch/none
    TaskEpilog              = (null)
    TaskPlugin              = task/none
    TaskPluginParam         = (null type)
    TaskProlog              = (null)
    TCPTimeout              = 2 sec
    TmpFS                   = /tmp
    TopologyParam           = (null)
    TopologyPlugin          = topology/none
    TrackWCKey              = No
    TreeWidth               = 50
    UsePam                  = No
    UnkillableStepProgram   = (null)
    UnkillableStepTimeout   = 60 sec
    VSizeFactor             = 0 percent
    WaitTime                = 0 sec
    X11Parameters           = (null)

    Slurmctld(primary) at democluster-example is UP
    """


@pytest.fixture
def parsed_scontrol_show_config() -> Generator[dict[str, str], None, None]:
    """Yield a parsed example of the `scontrol show config` command."""
    yield {
        "AccountingStorageBackupHost": "(null)",
        "AccountingStorageEnforce": "none",
        "AccountingStorageHost": "10.59.51.10",
        "AccountingStorageExternalHost": "(null)",
        "AccountingStorageParameters": "(null)",
        "AccountingStoragePort": "6839",
        "AccountingStorageTRES": "cpu,mem,energy,node,billing,fs/disk,vmem,pages",
        "AccountingStorageType": "accounting_storage/slurmdbd",
        "AccountingStorageUser": "N/A",
        "AccountingStoreFlags": "(null)",
        "AcctGatherEnergyType": "acct_gather_energy/none",
        "AcctGatherFilesystemType": "acct_gather_filesystem/none",
        "AcctGatherInterconnectType": "acct_gather_interconnect/none",
        "AcctGatherNodeFreq": "0 sec",
        "AcctGatherProfileType": "acct_gather_profile/none",
        "AllowSpecResourcesUsage": "No",
        "AuthAltTypes": "auth/jwt",
        "AuthAltParameters": "jwt_key=/var/lib/slurm/slurmctld/jwt_hs256.key",
        "AuthInfo": "socket=/var/run/munge/munge.socket.2",
        "AuthType": "auth/munge",
        "BatchStartTimeout": "10 sec",
        "BcastExclude": "/lib,/usr/lib,/lib64,/usr/lib64",
        "BcastParameters": "(null)",
        "BOOT_TIME": "2024-08-07T19:57:56",
        "BurstBufferType": "(null)",
        "CliFilterPlugins": "(null)",
        "ClusterName": "demo-cluster",
        "CommunicationParameters": "(null)",
        "CompleteWait": "0 sec",
        "CoreSpecPlugin": "core_spec/none",
        "CpuFreqDef": "Unknown",
        "CpuFreqGovernors": "OnDemand,Performance,UserSpace",
        "CredType": "cred/munge",
        "DebugFlags": "NO_CONF_HASH",
        "DefMemPerNode": "UNLIMITED",
        "DependencyParameters": "(null)",
        "DisableRootJobs": "No",
        "EioTimeout": "60",
        "EnforcePartLimits": "NO",
        "Epilog": "(null)",
        "EpilogMsgTime": "2000 usec",
        "EpilogSlurmctld": "(null)",
        "ExtSensorsType": "ext_sensors/none",
        "ExtSensorsFreq": "0 sec",
        "FederationParameters": "(null)",
        "FirstJobId": "1",
        "GetEnvTimeout": "2 sec",
        "GresTypes": "(null)",
        "GpuFreqDef": "(null)",
        "GroupUpdateForce": "1",
        "GroupUpdateTime": "600 sec",
        "HASH_VAL": "Match",
        "HealthCheckInterval": "0 sec",
        "HealthCheckNodeState": "ANY",
        "HealthCheckProgram": "(null)",
        "InactiveLimit": "0 sec",
        "InteractiveStepOptions": "--interactive --preserve-env --pty $SHELL",
        "JobAcctGatherFrequency": "30",
        "JobAcctGatherType": "jobacct_gather/linux",
        "JobAcctGatherParams": "(null)",
        "JobCompHost": "localhost",
        "JobCompLoc": "(null)",
        "JobCompParams": "(null)",
        "JobCompPort": "0",
        "JobCompType": "jobcomp/none",
        "JobCompUser": "root",
        "JobContainerType": "job_container/none",
        "JobCredentialPrivateKey": "(null)",
        "JobCredentialPublicCertificate": "(null)",
        "JobDefaults": "(null)",
        "JobFileAppend": "0",
        "JobRequeue": "1",
        "JobSubmitPlugins": "(null)",
        "KillOnBadExit": "0",
        "KillWait": "30 sec",
        "LaunchParameters": "(null)",
        "Licenses": "(null)",
        "LogTimeFormat": "iso8601_ms",
        "MailDomain": "(null)",
        "MailProg": "/bin/mail",
        "MaxArraySize": "1001",
        "MaxBatchRequeue": "5",
        "MaxDBDMsgs": "20004",
        "MaxJobCount": "10000",
        "MaxJobId": "67043328",
        "MaxMemPerNode": "UNLIMITED",
        "MaxNodeCount": "1",
        "MaxStepCount": "40000",
        "MaxTasksPerNode": "512",
        "MCSPlugin": "mcs/none",
        "MCSParameters": "(null)",
        "MessageTimeout": "10 sec",
        "MinJobAge": "300 sec",
        "MpiDefault": "none",
        "MpiParams": "(null)",
        "MULTIPLE_SLURMD": "Yes",
        "NEXT_JOB_ID": "1",
        "NodeFeaturesPlugins": "(null)",
        "OverTimeLimit": "0 min",
        "PluginDir": "/usr/lib/x86_64-linux-gnu/slurm-wlm/",
        "PlugStackConfig": "/etc/slurm/plugstack.conf",
        "PowerParameters": "(null)",
        "PowerPlugin": "",
        "PreemptMode": "OFF",
        "PreemptParameters": "(null)",
        "PreemptType": "preempt/none",
        "PreemptExemptTime": "00:00:00",
        "PrEpParameters": "(null)",
        "PrEpPlugins": "prep/script",
        "PriorityParameters": "(null)",
        "PrioritySiteFactorParameters": "(null)",
        "PrioritySiteFactorPlugin": "(null)",
        "PriorityType": "priority/basic",
        "PrivateData": "none",
        "ProctrackType": "proctrack/linuxproc",
        "Prolog": "(null)",
        "PrologEpilogTimeout": "65534",
        "PrologSlurmctld": "(null)",
        "PrologFlags": "(null)",
        "PropagatePrioProcess": "0",
        "PropagateResourceLimits": "ALL",
        "PropagateResourceLimitsExcept": "(null)",
        "RebootProgram": "(null)",
        "ReconfigFlags": "(null)",
        "RequeueExit": "(null)",
        "RequeueExitHold": "(null)",
        "ResumeFailProgram": "(null)",
        "ResumeProgram": "(null)",
        "ResumeRate": "300 nodes/min",
        "ResumeTimeout": "60 sec",
        "ResvEpilog": "(null)",
        "ResvOverRun": "0 min",
        "ResvProlog": "(null)",
        "ReturnToService": "2",
        "RoutePlugin": "route/default",
        "SchedulerParameters": "(null)",
        "SchedulerTimeSlice": "30 sec",
        "SchedulerType": "sched/backfill",
        "ScronParameters": "(null)",
        "SelectType": "select/cons_res",
        "SelectTypeParameters": "CR_CORE",
        "SlurmUser": "slurm(64030)",
        "SlurmctldAddr": "10.59.51.10",
        "SlurmctldDebug": "info",
        "SlurmctldHost[0]": "democluster-example",
        "SlurmctldLogFile": "/var/log/slurm/slurmctld.log",
        "SlurmctldPort": "6817",
        "SlurmctldSyslogDebug": "(null)",
        "SlurmctldPrimaryOffProg": "(null)",
        "SlurmctldPrimaryOnProg": "(null)",
        "SlurmctldTimeout": "300 sec",
        "SlurmctldParameters": "(null)",
        "SlurmdDebug": "info",
        "SlurmdLogFile": "/var/log/slurm/slurmd.log",
        "SlurmdParameters": "(null)",
        "SlurmdPidFile": "/var/run/slurmd.pid",
        "SlurmdPort": "6818",
        "SlurmdSpoolDir": "/var/lib/slurm/slurmd",
        "SlurmdSyslogDebug": "(null)",
        "SlurmdTimeout": "60 sec",
        "SlurmdUser": "root(0)",
        "SlurmSchedLogFile": "(null)",
        "SlurmSchedLogLevel": "0",
        "SlurmctldPidFile": "/var/run/slurmctld.pid",
        "SLURM_CONF": "/etc/slurm/slurm.conf",
        "SLURM_VERSION": "23.02.7",
        "SrunEpilog": "(null)",
        "SrunPortRange": "0-0",
        "SrunProlog": "(null)",
        "StateSaveLocation": "/var/lib/slurm/checkpoint",
        "SuspendExcNodes": "(null)",
        "SuspendExcParts": "(null)",
        "SuspendExcStates": "(null)",
        "SuspendProgram": "(null)",
        "SuspendRate": "60 nodes/min",
        "SuspendTime": "INFINITE",
        "SuspendTimeout": "30 sec",
        "SwitchParameters": "(null)",
        "SwitchType": "switch/none",
        "TaskEpilog": "(null)",
        "TaskPlugin": "task/none",
        "TaskPluginParam": "(null type)",
        "TaskProlog": "(null)",
        "TCPTimeout": "2 sec",
        "TmpFS": "/tmp",
        "TopologyParam": "(null)",
        "TopologyPlugin": "topology/none",
        "TrackWCKey": "No",
        "TreeWidth": "50",
        "UsePam": "No",
        "UnkillableStepProgram": "(null)",
        "UnkillableStepTimeout": "60 sec",
        "VSizeFactor": "0 percent",
        "WaitTime": "0 sec",
        "X11Parameters": "(null)",
    }


@pytest.fixture
def raw_scontrol_show_partition() -> Generator[str, None, None]:
    """Yield a raw example of the `scontrol show partition` command."""
    yield """
    PartitionName=debug
        AllowGroups=ALL AllowAccounts=ALL AllowQos=ALL
        AllocNodes=ALL Default=YES QoS=N/A
        DefaultTime=NONE DisableRootJobs=NO ExclusiveUser=NO GraceTime=0 Hidden=NO
        MaxNodes=3 MaxTime=60 MinNodes=0 LLN=NO MaxCPUsPerNode=32 MaxCPUsPerSocket=UNLIMITED
        Nodes=node[01-03]
        PriorityJobFactor=1 PriorityTier=1 RootOnly=NO ReqResv=NO OverSubscribe=NO
        OverTimeLimit=NONE PreemptMode=OFF
        State=UP TotalCPUs=96 TotalNodes=3 SelectTypeParameters=NONE
        JobDefaults=(null)
        DefMemPerNode=UNLIMITED MaxMemPerNode=UNLIMITED
        TRES=cpu=96,mem=UNLIMITED,node=3,billing=96

    PartitionName=normal
        AllowGroups=ALL AllowAccounts=ALL AllowQos=ALL
        AllocNodes=ALL Default=NO QoS=N/A
        DefaultTime=NONE DisableRootJobs=NO ExclusiveUser=NO GraceTime=0 Hidden=NO
        MaxNodes=7 MaxTime=INFINITE MinNodes=0 LLN=NO MaxCPUsPerNode=64 MaxCPUsPerSocket=UNLIMITED
        Nodes=node[04-10]
        PriorityJobFactor=1 PriorityTier=1 RootOnly=NO ReqResv=NO OverSubscribe=NO
        OverTimeLimit=NONE PreemptMode=OFF
        State=DOWN TotalCPUs=448 TotalNodes=7 SelectTypeParameters=NONE
        JobDefaults=(null)
        DefMemPerNode=UNLIMITED MaxMemPerNode=UNLIMITED
        TRES=cpu=448,mem=UNLIMITED,node=7,billing=448

    PartitionName=longrun
        AllowGroups=ALL AllowAccounts=ALL AllowQos=ALL
        AllocNodes=ALL Default=NO QoS=N/A
        DefaultTime=NONE DisableRootJobs=NO ExclusiveUser=NO GraceTime=0 Hidden=NO
        MaxNodes=5 MaxTime=INFINITE MinNodes=0 LLN=NO MaxCPUsPerNode=48 MaxCPUsPerSocket=UNLIMITED
        Nodes=node[11-15]
        PriorityJobFactor=1 PriorityTier=1 RootOnly=NO ReqResv=NO OverSubscribe=NO
        OverTimeLimit=NONE PreemptMode=OFF
        State=MAINT TotalCPUs=240 TotalNodes=5 SelectTypeParameters=NONE
        JobDefaults=(null)
        DefMemPerNode=UNLIMITED MaxMemPerNode=UNLIMITED
        TRES=cpu=240,mem=UNLIMITED,node=5,billing=240
    """


@pytest.fixture
def parsed_scontrol_show_partition() -> Generator[dict[str, dict[str, str]], None, None]:
    """Yield a parsed example of the `scontrol show partition` command."""
    yield {
        "debug": {
            "AllowGroups": "ALL",
            "AllowAccounts": "ALL",
            "AllowQos": "ALL",
            "AllocNodes": "ALL",
            "Default": "YES",
            "QoS": "N/A",
            "DefaultTime": "NONE",
            "DisableRootJobs": "NO",
            "ExclusiveUser": "NO",
            "GraceTime": "0",
            "Hidden": "NO",
            "MaxNodes": "3",
            "MaxTime": "60",
            "MinNodes": "0",
            "LLN": "NO",
            "MaxCPUsPerNode": "32",
            "MaxCPUsPerSocket": "UNLIMITED",
            "Nodes": "node[01-03]",
            "PriorityJobFactor": "1",
            "PriorityTier": "1",
            "RootOnly": "NO",
            "ReqResv": "NO",
            "OverSubscribe": "NO",
            "OverTimeLimit": "NONE",
            "PreemptMode": "OFF",
            "State": "UP",
            "TotalCPUs": "96",
            "TotalNodes": "3",
            "SelectTypeParameters": "NONE",
            "JobDefaults": "(null)",
            "DefMemPerNode": "UNLIMITED",
            "MaxMemPerNode": "UNLIMITED",
            "TRES": "cpu=96,mem=UNLIMITED,node=3,billing=96",
        },
        "normal": {
            "AllowGroups": "ALL",
            "AllowAccounts": "ALL",
            "AllowQos": "ALL",
            "AllocNodes": "ALL",
            "Default": "NO",
            "QoS": "N/A",
            "DefaultTime": "NONE",
            "DisableRootJobs": "NO",
            "ExclusiveUser": "NO",
            "GraceTime": "0",
            "Hidden": "NO",
            "MaxNodes": "7",
            "MaxTime": "INFINITE",
            "MinNodes": "0",
            "LLN": "NO",
            "MaxCPUsPerNode": "64",
            "MaxCPUsPerSocket": "UNLIMITED",
            "Nodes": "node[04-10]",
            "PriorityJobFactor": "1",
            "PriorityTier": "1",
            "RootOnly": "NO",
            "ReqResv": "NO",
            "OverSubscribe": "NO",
            "OverTimeLimit": "NONE",
            "PreemptMode": "OFF",
            "State": "DOWN",
            "TotalCPUs": "448",
            "TotalNodes": "7",
            "SelectTypeParameters": "NONE",
            "JobDefaults": "(null)",
            "DefMemPerNode": "UNLIMITED",
            "MaxMemPerNode": "UNLIMITED",
            "TRES": "cpu=448,mem=UNLIMITED,node=7,billing=448",
        },
        "longrun": {
            "AllowGroups": "ALL",
            "AllowAccounts": "ALL",
            "AllowQos": "ALL",
            "AllocNodes": "ALL",
            "Default": "NO",
            "QoS": "N/A",
            "DefaultTime": "NONE",
            "DisableRootJobs": "NO",
            "ExclusiveUser": "NO",
            "GraceTime": "0",
            "Hidden": "NO",
            "MaxNodes": "5",
            "MaxTime": "INFINITE",
            "MinNodes": "0",
            "LLN": "NO",
            "MaxCPUsPerNode": "48",
            "MaxCPUsPerSocket": "UNLIMITED",
            "Nodes": "node[11-15]",
            "PriorityJobFactor": "1",
            "PriorityTier": "1",
            "RootOnly": "NO",
            "ReqResv": "NO",
            "OverSubscribe": "NO",
            "OverTimeLimit": "NONE",
            "PreemptMode": "OFF",
            "State": "MAINT",
            "TotalCPUs": "240",
            "TotalNodes": "5",
            "SelectTypeParameters": "NONE",
            "JobDefaults": "(null)",
            "DefMemPerNode": "UNLIMITED",
            "MaxMemPerNode": "UNLIMITED",
            "TRES": "cpu=240,mem=UNLIMITED,node=5,billing=240",
        },
    }


@pytest.fixture
def raw_scontrol_show_node() -> Generator[str, None, None]:
    """Yield a raw example of the `scontrol show node` command."""
    yield """
    NodeName=node01 Arch=x86_64 CoresPerSocket=8
        CPUAlloc=2 CPUErr=0 CPUTot=16 CPULoad=0.05
        AvailableFeatures=(null) ActiveFeatures=(null)
        Gres=gpu:2 NodeAddr=node01 NodeHostName=node01 Version=20.11
        OS=Linux 3.10.0-1160.el7.x86_64 #1 SMP Tue Sep 22 17:24:55 UTC 2020
        RealMemory=64000 AllocMem=20000 FreeMem=40000 Sockets=2 Boards=1
        State=IDLE ThreadsPerCore=2 TmpDisk=0 Weight=1 Owner=N/A MCS_label=N/A
        BootTime=2021-04-15T10:05:55 SlurmdStartTime=2021-04-15T10:10:30
        CfgTRES=cpu=16,mem=64000M,billing=16
        AllocTRES=cpu=2,mem=20000M
        CapWatts=n/a
        CurrentWatts=0 LowestJoules=0 ConsumedJoules=0
        ExtSensorsJoules=n/s ExtSensorsWatts=0 ExtSensorsTemp=n/s
        Reason=NHC: check_ps_service:  Service sshd (process sshd) owned by root not running [root@2024-08-20T16:32:47]

    NodeName=node02 Arch=x86_64 CoresPerSocket=8
        CPUAlloc=0 CPUErr=0 CPUTot=16 CPULoad=0.00
        AvailableFeatures=(null) ActiveFeatures=(null)
        Gres=gpu:2 NodeAddr=node02 NodeHostName=node02 Version=20.11
        OS=Linux 3.10.0-1160.el7.x86_64 #1 SMP Tue Sep 22 17:24:55 UTC 2020
        RealMemory=64000 AllocMem=0 FreeMem=64000 Sockets=2 Boards=1
        State=ALLOCATED ThreadsPerCore=2 TmpDisk=0 Weight=1 Owner=N/A MCS_label=N/A
        BootTime=2021-04-15T10:05:55 SlurmdStartTime=2021-04-15T10:10:30
        CfgTRES=cpu=16,mem=64000M,billing=16
        AllocTRES=cpu=0,mem=0M
        CapWatts=n/a
        CurrentWatts=0 LowestJoules=0 ConsumedJoules=0
        ExtSensorsJoules=n/s ExtSensorsWatts=0 ExtSensorsTemp=n/s
        Reason=NHC: check_ps_service:  Service sshd (process sshd) owned by root not running [root@2024-08-20T16:32:47]
    """


@pytest.fixture
def raw_squeue_all() -> Generator[str, None, None]:
    """Yield a raw example of the `squeue -o %all` command."""
    yield """
    ACCOUNT|TRES_PER_NODE|MIN_CPUS|MIN_TMP_DISK|END_TIME|FEATURES|GROUP|OVER_SUBSCRIBE|JOBID|NAME|COMMENT|TIME_LIMIT|MIN_MEMORY|REQ_NODES|COMMAND|PRIORITY|QOS|REASON|ST|USER|RESERVATION|WCKEY|EXC_NODES|NICE|S:C:T|JOBID|EXEC_HOST|CPUS|NODES|DEPENDENCY|ARRAY_JOB_ID|GROUP|SOCKETS_PER_NODE|CORES_PER_SOCKET|THREADS_PER_CORE|ARRAY_TASK_ID|TIME_LEFT|TIME|NODELIST|CONTIGUOUS|PARTITION|PRIORITY|NODELIST(REASON)|START_TIME|STATE|UID|SUBMIT_TIME|LICENSES|CORE_SPEC|SCHEDNODES|WORK_DIR
(null)|N/A|1|0|NONE|(null)|ubuntu|OK|1|spawner-jupyterhub|(null)|UNLIMITED|0||(null)|0.99998474121093|normal|None|R|ubuntu|(null)|(null)||0|*:*:*|1|compute-compute-4|1|1|(null)|1|1000|*|*|*|N/A|UNLIMITED|15:56:32|compute-compute-4|0|compute|4294901759|compute-compute-4|2025-06-17T00:55:01|RUNNING|1000|2025-06-17T00:51:03|(null)|N/A|(null)|/home/ubuntu
(null)|N/A|1|0|NONE|(null)|ubuntu|OK|2|spawner-jupyterhub|(null)|UNLIMITED|0||(null)|0.99998474097810|normal|None|R|ubuntu|(null)|(null)||0|*:*:*|2|compute-compute-4|1|1|(null)|2|1000|*|*|*|N/A|UNLIMITED|15:56:32|compute-compute-4|0|compute|4294901758|compute-compute-4|2025-06-17T00:55:01|RUNNING|1000|2025-06-17T00:51:04|(null)|N/A|(null)|/home/ubuntu
    """  # noqa


@pytest.fixture
def parsed_squeue_all() -> Generator[list[dict[str, str]], None, None]:
    """Yield a parsed example of the `squeue -o %all` command."""
    yield {
        "1-spawner-jupyterhub": {
            "account": "(null)",
            "tres_per_node": "N/A",
            "min_cpus": "1",
            "min_tmp_disk": "0",
            "end_time": "NONE",
            "features": "(null)",
            "group": "1000",
            "over_subscribe": "OK",
            "jobid": "1",
            "name": "spawner-jupyterhub",
            "comment": "(null)",
            "time_limit": "UNLIMITED",
            "min_memory": "0",
            "req_nodes": "",
            "command": "(null)",
            "priority": "4294901759",
            "qos": "normal",
            "reason": "None",
            "st": "R",
            "user": "ubuntu",
            "reservation": "(null)",
            "wckey": "(null)",
            "exc_nodes": "",
            "nice": "0",
            "s:c:t": "*:*:*",
            "exec_host": "compute-compute-4",
            "cpus": "1",
            "nodes": "1",
            "dependency": "(null)",
            "array_job_id": "1",
            "sockets_per_node": "*",
            "cores_per_socket": "*",
            "threads_per_core": "*",
            "array_task_id": "N/A",
            "time_left": "UNLIMITED",
            "time": "15:56:32",
            "nodelist": "compute-compute-4",
            "contiguous": "0",
            "partition": "compute",
            "nodelist(reason)": "compute-compute-4",
            "start_time": "2025-06-17T00:55:01",
            "state": "RUNNING",
            "uid": "1000",
            "submit_time": "2025-06-17T00:51:03",
            "licenses": "(null)",
            "core_spec": "N/A",
            "schednodes": "(null)",
            "work_dir": "/home/ubuntu",
        },
        "2-spawner-jupyterhub": {
            "account": "(null)",
            "tres_per_node": "N/A",
            "min_cpus": "1",
            "min_tmp_disk": "0",
            "end_time": "NONE",
            "features": "(null)",
            "group": "1000",
            "over_subscribe": "OK",
            "jobid": "2",
            "name": "spawner-jupyterhub",
            "comment": "(null)",
            "time_limit": "UNLIMITED",
            "min_memory": "0",
            "req_nodes": "",
            "command": "(null)",
            "priority": "4294901758",
            "qos": "normal",
            "reason": "None",
            "st": "R",
            "user": "ubuntu",
            "reservation": "(null)",
            "wckey": "(null)",
            "exc_nodes": "",
            "nice": "0",
            "s:c:t": "*:*:*",
            "exec_host": "compute-compute-4",
            "cpus": "1",
            "nodes": "1",
            "dependency": "(null)",
            "array_job_id": "2",
            "sockets_per_node": "*",
            "cores_per_socket": "*",
            "threads_per_core": "*",
            "array_task_id": "N/A",
            "time_left": "UNLIMITED",
            "time": "15:56:32",
            "nodelist": "compute-compute-4",
            "contiguous": "0",
            "partition": "compute",
            "nodelist(reason)": "compute-compute-4",
            "start_time": "2025-06-17T00:55:01",
            "state": "RUNNING",
            "uid": "1000",
            "submit_time": "2025-06-17T00:51:04",
            "licenses": "(null)",
            "core_spec": "N/A",
            "schednodes": "(null)",
            "work_dir": "/home/ubuntu",
        },
    }


@pytest.fixture
def parsed_scontrol_show_node() -> Generator[dict[str, dict[str, str]], None, None]:
    """Yield a parsed example of the `scontrol show node` command."""
    yield {
        "node01": {
            "Arch": "x86_64",
            "CoresPerSocket": "8",
            "CPUAlloc": "2",
            "CPUErr": "0",
            "CPUTot": "16",
            "CPULoad": "0.05",
            "AvailableFeatures": "(null)",
            "ActiveFeatures": "(null)",
            "Gres": "gpu:2",
            "NodeAddr": "node01",
            "NodeHostName": "node01",
            "Version": "20.11",
            "OS": "Linux 3.10.0-1160.el7.x86_64 #1 SMP Tue Sep 22 17:24:55 UTC 2020",
            "RealMemory": "64000",
            "AllocMem": "20000",
            "FreeMem": "40000",
            "Sockets": "2",
            "Boards": "1",
            "State": "IDLE",
            "ThreadsPerCore": "2",
            "TmpDisk": "0",
            "Weight": "1",
            "Owner": "N/A",
            "MCS_label": "N/A",
            "BootTime": "2021-04-15T10:05:55",
            "SlurmdStartTime": "2021-04-15T10:10:30",
            "CfgTRES": "cpu=16,mem=64000M,billing=16",
            "AllocTRES": "cpu=2,mem=20000M",
            "CapWatts": "n/a",
            "CurrentWatts": "0",
            "LowestJoules": "0",
            "ConsumedJoules": "0",
            "ExtSensorsJoules": "n/s",
            "ExtSensorsWatts": "0",
            "ExtSensorsTemp": "n/s",
            "Reason": "NHC: check_ps_service:  Service sshd (process sshd) owned by root not running [root@2024-08-20T16:32:47]",
        },
        "node02": {
            "Arch": "x86_64",
            "CoresPerSocket": "8",
            "CPUAlloc": "0",
            "CPUErr": "0",
            "CPUTot": "16",
            "CPULoad": "0.00",
            "AvailableFeatures": "(null)",
            "ActiveFeatures": "(null)",
            "Gres": "gpu:2",
            "NodeAddr": "node02",
            "NodeHostName": "node02",
            "Version": "20.11",
            "OS": "Linux 3.10.0-1160.el7.x86_64 #1 SMP Tue Sep 22 17:24:55 UTC 2020",
            "RealMemory": "64000",
            "AllocMem": "0",
            "FreeMem": "64000",
            "Sockets": "2",
            "Boards": "1",
            "State": "ALLOCATED",
            "ThreadsPerCore": "2",
            "TmpDisk": "0",
            "Weight": "1",
            "Owner": "N/A",
            "MCS_label": "N/A",
            "BootTime": "2021-04-15T10:05:55",
            "SlurmdStartTime": "2021-04-15T10:10:30",
            "CfgTRES": "cpu=16,mem=64000M,billing=16",
            "AllocTRES": "cpu=0,mem=0M",
            "CapWatts": "n/a",
            "CurrentWatts": "0",
            "LowestJoules": "0",
            "ConsumedJoules": "0",
            "ExtSensorsJoules": "n/s",
            "ExtSensorsWatts": "0",
            "ExtSensorsTemp": "n/s",
            "Reason": "NHC: check_ps_service:  Service sshd (process sshd) owned by root not running [root@2024-08-20T16:32:47]",
        },
    }


class TestParseCorrectSlurmOutput:
    """Test the slurm output parsing functions when the output is pretty."""

    def test_parse_slurm_config(self, raw_scontrol_show_config, parsed_scontrol_show_config):
        """Test the `parse_slurm_config` function."""
        assert parse_slurm_config(raw_scontrol_show_config) == parsed_scontrol_show_config

    def test_parse_scontrol_show_partition(self, raw_scontrol_show_partition, parsed_scontrol_show_partition):
        """Test the `parse_scontrol_show_partition` function."""
        assert parse_slurm_partitions(raw_scontrol_show_partition) == parsed_scontrol_show_partition

    def test_parse_squeue_all(self, raw_squeue_all, parsed_squeue_all):
        """Test the `parse_squeue_all` function."""
        assert parse_slurm_queue(raw_squeue_all) == parsed_squeue_all

    def test_parse_scontrol_show_node(self, raw_scontrol_show_node, parsed_scontrol_show_node):
        """Test the `parse_scontrol_show_node` function."""
        assert parse_slurm_nodes(raw_scontrol_show_node) == parsed_scontrol_show_node


class TestParseBuggySlurmOutput:
    """Test the slurm output parsing functions when the output is buggy."""

    def test_parse_buggy_slurm_config(self):
        """Test parsing the `scontrol show config` command when the output is buggy."""
        raw_output = """
        AccountingStorageBackupHost=(null)
        AccountingStorageEnforce=none
        AccountingStorageHost-buggy
        """

        expected_output = {
            "AccountingStorageBackupHost": "(null)",
            "AccountingStorageEnforce": "none",
        }

        assert parse_slurm_config(raw_output) == expected_output

    def test_parse_buggy_scontrol_show_partition(self):
        """Test parsing the `scontrol show partition` command when the output is buggy."""
        raw_output = """
        PartitionName=debug
            AllocNodes$$$$ALL Default(#&)YES MaxNodes=3 OverSubscribe!!--NO
            QoS=buggy
        """

        expected_output = {
            "debug": {
                "MaxNodes": "3",
                "QoS": "buggy",
            }
        }

        assert parse_slurm_partitions(raw_output) == expected_output

    def test_parse_empty_squeue_all(self):
        """Test parsing the `squeue -o %all` command when the output is empty."""
        raw_output = """
        ACCOUNT|TRES_PER_NODE|MIN_CPUS|MIN_TMP_DISK|END_TIME|FEATURES|GROUP|OVER_SUBSCRIBE|JOBID|NAME|COMMENT|TIME_LIMIT|MIN_MEMORY|REQ_NODES|COMMAND|PRIORITY|QOS|REASON|ST|USER|RESERVATION|WCKEY|EXC_NODES|NICE|S:C:T|JOBID|EXEC_HOST|CPUS|NODES|DEPENDENCY|ARRAY_JOB_ID|GROUP|SOCKETS_PER_NODE|CORES_PER_SOCKET|THREADS_PER_CORE|ARRAY_TASK_ID|TIME_LEFT|TIME|NODELIST|CONTIGUOUS|PARTITION|PRIORITY|NODELIST(REASON)|START_TIME|STATE|UID|SUBMIT_TIME|LICENSES|CORE_SPEC|SCHEDNODES|WORK_DIR
        """

        expected_output = {}

        assert parse_slurm_queue(raw_output) == expected_output

    def test_parse_buggy_squeue_all(self):
        """Test parsing the `squeue -o %all` command when the output is empty."""
        raw_output = """
        ACCOUNT|TRES_PER_NODE|MIN_CPUS|MIN_TMP_DISK|END_TIME|FEATURES|GROUP|OVER_SUBSCRIBE|JOBID|NAME
        (null)|N/A|1|0|NONE|(null)|ubuntu|OK|1|spawner-jupyterhub|(null)|UNLIMITED|0||(null)
        """

        expected_output = {
            "1-spawner-jupyterhub": {
                "account": "(null)",
                "tres_per_node": "N/A",
                "min_cpus": "1",
                "min_tmp_disk": "0",
                "end_time": "NONE",
                "features": "(null)",
                "group": "ubuntu",
                "over_subscribe": "OK",
                "jobid": "1",
                "name": "spawner-jupyterhub",
            }
        }

        assert parse_slurm_queue(raw_output) == expected_output

        raw_output = """
        ACCOUNT|TRES_PER_NODE|MIN_CPUS|MIN_TMP_DISK|END_TIME|FEATURES|GROUP|OVER_SUBSCRIBE|JOBID|NAME|
        (null)|N/A|1|0|(null)|ubuntu|OK|1|spawner-jupyterhub
        """

        expected_output = {
            "spawner-jupyterhub-": {
                "account": "(null)",
                "tres_per_node": "N/A",
                "min_cpus": "1",
                "min_tmp_disk": "0",
                "end_time": "(null)",
                "features": "ubuntu",
                "group": "OK",
                "over_subscribe": "1",
                "jobid": "spawner-jupyterhub",
                "name": "",
                "": "",
            }
        }

        assert parse_slurm_queue(raw_output) == expected_output

    def test_parse_buggy_scontrol_show_node(self):
        """Test parsing the `scontrol show node` command when the output is buggy."""
        raw_output = """
        NodeName=buggy Arch=x86_64 CoresPerSocket*$&#8
            CPUAlloc=0 CPUErr!!--$%0 CPUTot=16 CPULoad###0.00
        """

        expected_output = {
            "buggy": {
                "Arch": "x86_64",
                "CPUAlloc": "0",
                "CPUTot": "16",
            }
        }

        assert parse_slurm_nodes(raw_output) == expected_output


class TestLoadCachedDict:
    """Test the `load_cached_dict` function."""

    @pytest.mark.parametrize(
        "file_name, expected_data",
        [
            ("foo.json", {"alpha": "beta", "gamma": "delta"}),
            ("boo.json", {"epsilon": "zeta", "eta": "theta"}),
        ],
    )
    def test_load_existing_json(
        self, file_name: str, expected_data: dict[str, str], tmp_path_factory: pytest.TempPathFactory
    ):
        """Test loading an existing JSON file."""
        cache_dir: Path = tmp_path_factory.mktemp("cache")

        with open(cache_dir / f"{file_name}", "w") as f:
            json.dump(expected_data, f)

        with mock.patch("vantage_agent.helpers.SETTINGS.CACHE_DIR", str(cache_dir)):
            result: dict[str, str] = load_cached_dict(file_name)

        assert result == expected_data

    @pytest.mark.parametrize("file_name", ["baz", "qux"])
    def test_load_nonexistent_file(self, file_name: str, tmp_path_factory: pytest.TempPathFactory):
        """Test loading a non-existent file."""
        cache_dir: Path = tmp_path_factory.mktemp("cache")

        # Patch SETTINGS.CACHE_DIR and call the function
        with mock.patch("vantage_agent.helpers.SETTINGS.CACHE_DIR", str(cache_dir)):
            result = load_cached_dict(file_name)

        assert result == {}

    def test_load_invalid_json(self, tmp_path_factory: pytest.TempPathFactory):
        """Test loading an invalid JSON file."""
        file_name = "invalid_data"
        cache_dir: Path = tmp_path_factory.mktemp("cache")

        with open(f"{cache_dir}/{file_name}.json", "w") as f:
            f.write("invalid json")

        with mock.patch("vantage_agent.helpers.SETTINGS.CACHE_DIR", str(cache_dir)):
            result = load_cached_dict(file_name)

        assert result == {}

    @pytest.mark.parametrize(
        "file_name, expected_data",
        [
            ("baz", {"iota": "kappa", "lambda": "mu"}),
            ("qux", {"nu": "xi", "omicron": "pi"}),
        ],
    )
    def test_file_name_without_extension(
        self, file_name: str, expected_data: dict[str, str], tmp_path_factory: pytest.TempPathFactory
    ):
        """Test providing a file name without the .json extension."""
        cache_dir: Path = tmp_path_factory.mktemp("cache")

        with open(f"{cache_dir}/{file_name}.json", "w") as f:
            json.dump(expected_data, f)

        with mock.patch("vantage_agent.helpers.SETTINGS.CACHE_DIR", str(cache_dir)):
            result = load_cached_dict(file_name)

        assert result == expected_data


class TestCacheDict:
    """Test the `cache_dict` function."""

    @pytest.mark.parametrize(
        "file_name, content",
        [
            ("baz", {"lakers": "seahawks", "steelers": "falcons"}),
            ("qux", {"saints": "bulls", "raptors": "packers"}),
        ],
    )
    def test_cache_valid_dict(
        self, file_name: str, content: dict[str, str], tmp_path_factory: pytest.TempPathFactory
    ):
        """Test caching a valid dictionary."""
        cache_dir: Path = tmp_path_factory.mktemp("cache")

        with mock.patch("vantage_agent.helpers.SETTINGS.CACHE_DIR", str(cache_dir)):
            cache_dict(content, file_name)

        cached_file_path = cache_dir / f"{file_name}.json"
        assert cached_file_path.exists()

        with open(cached_file_path, "r") as f:
            cached_data = json.load(f)

        assert cached_data == content

    @pytest.mark.parametrize(
        "file_name, content",
        [
            ("zoo", {"eagles": "bengals", "patriots": "london"}),
            ("bar", {"broncos": "raiders", "chargers": "seattle"}),
        ],
    )
    def test_file_name_without_extension(
        self, file_name: str, content: dict[str, str], tmp_path_factory: pytest.TempPathFactory
    ):
        """Test providing a file name without the .json extension."""
        cache_dir: Path = tmp_path_factory.mktemp("cache")

        with mock.patch("vantage_agent.helpers.SETTINGS.CACHE_DIR", str(cache_dir)):
            cache_dict(content, file_name)

        cached_file_path = cache_dir / f"{file_name}.json"
        assert cached_file_path.exists()

    def test_override_existing_cache(self, tmp_path_factory: pytest.TempPathFactory):
        """Test overriding an existing cached dictionary."""
        initial_dict = {"key1": "value1"}
        new_dict = {"key2": "value2"}
        file_name = "test_override"
        cache_dir: Path = tmp_path_factory.mktemp("cache")

        with open(cache_dir / f"{file_name}.json", "w") as f:
            json.dump(initial_dict, f)

        with mock.patch("vantage_agent.helpers.SETTINGS.CACHE_DIR", str(cache_dir)):
            cache_dict(new_dict, file_name)

        cached_file_path = cache_dir / f"{file_name}.json"
        assert cached_file_path.exists()

        with open(cached_file_path, "r") as f:
            cached_data = json.load(f)

        assert cached_data == new_dict



"""Core module for defining the self update logic for the Vantage Agent."""

import re
import subprocess
import sys
from importlib.metadata import version

from apscheduler.job import Job
from loguru import logger

from vantage_agent.scheduler import schedule_tasks, scheduler
from vantage_agent.vantage_api_client import AsyncBackendClient

package_name = "vantage_agent"


async def _fetch_upstream_version_info() -> str:
    logger.debug("Fetching the upstream version info from vantage API")
    async with AsyncBackendClient() as backend_client:
        response = await backend_client.get("/cluster/openapi.json")
    response.raise_for_status()

    data = response.json()
    upstream_version: str = data["info"]["version"]
    return upstream_version


def _need_update(current_version: str, upstream_version: str) -> bool:
    """Compare the current version with the upstream version.

    In case the current version is the same as the upstream version, return False.
    If the major versions are different, return False, as updates across major versions are not desired.
    Otherwise, return True, allowing updates and rollbacks across all minor and patch versions,
    including handling for pre-release versions in case the release is tagged with `a` as alpha,
    e.g. `1.0.0a1`.
    """
    current_major: int | str
    current_minor: int | str
    current_patch: int | str
    upstream_major: int | str
    upstream_minor: int | str
    upstream_patch: int | str

    # regular expression to parse version strings: major.minor.patch
    version_pattern = r"^(\d+)\.(\d+)\.(\d+)(?:-(alpha)\.?(\d+)|a(\d+))?$"

    current_match = re.match(version_pattern, current_version)
    upstream_match = re.match(version_pattern, upstream_version)

    if not current_match or not upstream_match:
        raise ValueError(
            f"One of the following versions are improperly formatted: {current_version}, {upstream_version}"
        )

    (
        current_major,
        current_minor,
        current_patch,
        current_prerelease_type,
        current_prerelease_ver,
        current_alpha,
    ) = current_match.groups()
    (
        upstream_major,
        upstream_minor,
        upstream_patch,
        upstream_prerelease_type,
        upstream_prerelease_ver,
        upstream_alpha,
    ) = upstream_match.groups()

    if current_prerelease_type:
        current_alpha = current_prerelease_ver
    if upstream_prerelease_type:
        upstream_alpha = upstream_prerelease_ver

    current_major, current_minor, current_patch, current_alpha = map(
        int, [current_major, current_minor, current_patch, current_alpha or 0]
    )
    upstream_major, upstream_minor, upstream_patch, upstream_alpha = map(
        int, [upstream_major, upstream_minor, upstream_patch, upstream_alpha or 0]
    )

    # check tests/test_logicals/test_update.py for understanding the below logic
    if current_major != upstream_major:
        return False
    elif current_minor != upstream_minor:
        if upstream_alpha != 0:
            return False
        return True
    elif current_patch != upstream_patch:
        # don't rollback current alpha version
        if current_alpha != 0 and current_patch > upstream_patch:
            return False
        return True
    elif (upstream_alpha == 0 and current_alpha != 0) or upstream_alpha > current_alpha:
        return True
    return False


def _update_package(version: str) -> None:
    subprocess.check_call([sys.executable, "-m", "pip", "install", "--upgrade", f"{package_name}=={version}"])


async def self_update_agent() -> None:
    """Fetch the upstream version and update the agent if necessary.

    In case the agent is updated, the scheduler is shutdown and restarted with the new version.
    """
    current_version = version(package_name)
    upstream_version = await _fetch_upstream_version_info()
    logger.debug(
        f"Vantage Agent version info: current_version={current_version}, upstream_version={upstream_version}"
    )

    if _need_update(current_version, upstream_version):
        logger.warning(
            "The Vantage Agent is outdated in relation to the upstream version; an update is required."
        )

        logger.debug("Pausing the scheduler...")
        scheduler.pause()

        logger.debug("Clearing the scheduler jobs...")
        scheduler_jobs: list[Job] = scheduler.get_jobs()
        for job in scheduler_jobs:
            job.remove()

        logger.debug(f"Updating {package_name} from version {current_version} to {upstream_version}...")
        _update_package(upstream_version)
        logger.debug("Update completed successfully.")

        logger.debug(f"Loading plugins from version {upstream_version}...")
        schedule_tasks(scheduler)
        logger.debug("Plugins loaded successfully.")

        logger.debug("Resuming the scheduler...")
        scheduler.resume()
        logger.debug("Scheduler resumed successfully.")
    else:
        logger.debug("No update is required or update crosses a major version divide.")



"""Core module for testing the vantage_agent/logicals/vantage_api.py module."""

from textwrap import dedent
from typing import Generator
from unittest import mock

import pytest
from _pytest.monkeypatch import MonkeyPatch
from respx.router import MockRouter

from vantage_agent.logicals.vantage_api import (
    clean_slurm_nodes,
    clean_slurm_partitions,
    upload_slurm_config,
    upload_slurm_nodes,
    upload_slurm_partitions,
    upload_slurm_queue,
)
from vantage_agent.vantage_api_client import backend_client


@pytest.fixture
def mock_logger(monkeypatch: MonkeyPatch) -> Generator[mock.MagicMock, None, None]:
    """Mock the logger."""
    mock_logger = mock.Mock()
    monkeypatch.setattr("vantage_agent.logicals.vantage_api.logger", mock_logger)
    return mock_logger


@pytest.fixture
def mock_settings(monkeypatch: MonkeyPatch) -> Generator[mock.MagicMock, None, None]:
    """Mock the settings."""
    mock_settings = mock.Mock()
    monkeypatch.setattr("vantage_agent.logicals.vantage_api.SETTINGS", mock_settings)
    return mock_settings


@pytest.mark.respx(base_url=backend_client.base_url)
@pytest.mark.usefixtures("mock_access_token")
@pytest.mark.parametrize(
    "client_id, slurm_config",
    [
        ("madrid", {"bangkok": "bali"}),
        ("barcelona", {"cancun": "nairobi"}),
        ("valencia", {"helsinki": "copenhagen"}),
    ],
)
async def test_upload_slurm_config(
    client_id: str,
    slurm_config: dict[str, str],
    respx_mock: MockRouter,
    mock_settings: mock.MagicMock,
    mock_logger: mock.MagicMock,
):
    """Test if the upload_slurm_config send the proper GraphQL query to the Vantage API."""
    mock_settings.OIDC_CLIENT_ID = client_id
    mock_settings.BASE_API_URL = backend_client.base_url

    response_payload = {
        "data": {"UploadSlurmConfigSuccess": {"message": "Slurm config uploaded successfully"}}
    }

    query = dedent(
        """
        mutation($config: JSONScalar!, $clientId: String!) {
            uploadSlurmConfig(config: $config, clientId: $clientId) {
                ... on UploadSlurmConfigSuccess {
                    message
                }
            }
        }
        """
    )
    body = {"query": query, "variables": {"config": slurm_config, "clientId": client_id}}

    respx_mock.post("/cluster/graphql", json=body).respond(status_code=200, json=response_payload)

    await upload_slurm_config(slurm_config)

    mock_logger.debug.assert_any_call(
        f"Response from Vantage API while uploading Slurm's configurations: {response_payload}"
    )


@pytest.mark.respx(base_url=backend_client.base_url)
@pytest.mark.usefixtures("mock_access_token")
@pytest.mark.parametrize(
    "client_id, slurm_nodes",
    [
        ("tokyo", {"berlin": "frankfurt"}),
        ("istanbul", {"cape town": "cairo"}),
        ("mumbai", {"santiago": "paris"}),
    ],
)
async def test_upload_slurm_nodes(
    client_id: str,
    slurm_nodes: dict[str, str],
    respx_mock: MockRouter,
    mock_settings: mock.MagicMock,
    mock_logger: mock.MagicMock,
):
    """Test if the upload_slurm_nodes send the proper GraphQL query to the Vantage API."""
    mock_settings.OIDC_CLIENT_ID = client_id
    mock_settings.BASE_API_URL = backend_client.base_url

    response_payload = {"data": {"UploadSlurmNodesSuccess": {"message": "Slurm nodes uploaded successfully"}}}

    query = dedent(
        """
        mutation($nodes: JSONScalar!, $clientId: String!) {
            uploadSlurmNodes(nodes: $nodes, clientId: $clientId) {
                ... on UploadSlurmNodesSuccess {
                    message
                }
            }
        }
        """
    )
    body = {"query": query, "variables": {"nodes": slurm_nodes, "clientId": client_id}}

    respx_mock.post("/cluster/graphql", json=body).respond(status_code=200, json=response_payload)

    await upload_slurm_nodes(slurm_nodes)

    mock_logger.debug.assert_any_call(f"Response from Vantage API while uploading nodes: {response_payload}")


@pytest.mark.respx(base_url=backend_client.base_url)
@pytest.mark.usefixtures("mock_access_token")
@pytest.mark.parametrize(
    "client_id, slurm_partitions",
    [
        ("amsterdam", {"stockholm": "jakarta"}),
        ("lagos", {"new delhi": "seoul"}),
        ("lima", {"los angeles": "kolkata"}),
    ],
)
async def test_upload_slurm_partitions(
    client_id: str,
    slurm_partitions: dict[str, str],
    respx_mock: MockRouter,
    mock_settings: mock.MagicMock,
    mock_logger: mock.MagicMock,
):
    """Test if the upload_slurm_partitions send the proper GraphQL query to the Vantage API."""
    mock_settings.OIDC_CLIENT_ID = client_id
    mock_settings.BASE_API_URL = backend_client.base_url

    response_payload = {"data": {"uploadSlurmPartitions": {"message": "Slurm config uploaded successfully"}}}

    query = dedent(
        """
        mutation($partitions: JSONScalar!, $clientId: String!) {
            uploadSlurmPartitions(partitions: $partitions, clientId: $clientId) {
                ... on UploadSlurmPartitionsSuccess {
                    message
                }
            }
        }
        """
    )
    body = {"query": query, "variables": {"partitions": slurm_partitions, "clientId": client_id}}

    respx_mock.post("/cluster/graphql", json=body).respond(status_code=200, json=response_payload)

    await upload_slurm_partitions(slurm_partitions)

    mock_logger.debug.assert_any_call(
        f"Response from Vantage API while uploading partitions: {response_payload}"
    )


@pytest.mark.respx(base_url=backend_client.base_url)
@pytest.mark.usefixtures("mock_access_token")
@pytest.mark.parametrize(
    "client_id, slurm_partitions",
    [
        ("ancara", ["partition1", "partition2"]),
        ("manaus", ["partition3"]),
    ],
)
async def test_clean_slurm_partitions(
    client_id: str,
    slurm_partitions: list[str],
    respx_mock: MockRouter,
    mock_settings: mock.MagicMock,
    mock_logger: mock.MagicMock,
):
    """Test if clean_slurm_partitions sends the proper GraphQL query."""
    mock_settings.OIDC_CLIENT_ID = client_id
    mock_settings.BASE_API_URL = backend_client.base_url

    response_payload = {
        "data": {"CleanSlurmPartitionsSuccess": {"message": "Slurm partitions cleaned successfully"}}
    }

    query = dedent(
        """
        mutation($partitions: [String!]!, $clientId: String!) {
            cleanSlurmPartitions(partitions: $partitions, clientId: $clientId) {
                ... on CleanSlurmPartitionsSuccess {
                    message
                }
            }
        }
        """
    )
    body = {"query": query, "variables": {"partitions": slurm_partitions, "clientId": client_id}}

    respx_mock.post("/cluster/graphql", json=body).respond(status_code=200, json=response_payload)

    await clean_slurm_partitions(slurm_partitions)

    mock_logger.debug.assert_any_call(
        f"Response from Vantage API while cleaning partitions: {response_payload}"
    )


@pytest.mark.respx(base_url=backend_client.base_url)
@pytest.mark.usefixtures("mock_access_token")
@pytest.mark.parametrize(
    "client_id, slurm_nodes",
    [
        ("quito", ["node1", "node2"]),
        ("caracas", ["node3"]),
    ],
)
async def test_clean_slurm_nodes(
    client_id: str,
    slurm_nodes: list[str],
    respx_mock: MockRouter,
    mock_settings: mock.MagicMock,
    mock_logger: mock.MagicMock,
):
    """Test if clean_slurm_nodes sends the proper GraphQL query."""
    mock_settings.OIDC_CLIENT_ID = client_id
    mock_settings.BASE_API_URL = backend_client.base_url

    response_payload = {"data": {"CleanSlurmNodesSuccess": {"message": "Slurm nodes cleaned successfully"}}}

    query = dedent(
        """
        mutation($nodes: [String!]!, $clientId: String!) {
            cleanSlurmNodes(nodes: $nodes, clientId: $clientId) {
                ... on CleanSlurmNodesSuccess {
                    message
                }
            }
        }
        """
    )
    body = {"query": query, "variables": {"nodes": slurm_nodes, "clientId": client_id}}

    respx_mock.post("/cluster/graphql", json=body).respond(status_code=200, json=response_payload)

    await clean_slurm_nodes(slurm_nodes)

    mock_logger.debug.assert_any_call(f"Response from Vantage API while cleaning nodes: {response_payload}")


@pytest.mark.respx(base_url=backend_client.base_url)
@pytest.mark.usefixtures("mock_access_token")
@pytest.mark.parametrize(
    "client_id, slurm_queue",
    [
        (
            "clusterA-39bb8c7a-f882-4374-93ad-762d65ed06a1",
            [{"name": "job-test", "state": "CONFIGURING"}, {"name": "job1", "state": "RUNNING"}],
        ),
        ("clusterB-39bb8c7a-f882-4374-93ad-762d65ed06a1", [{"name": "second-job", "state": "PENDING"}]),
    ],
)
async def test_upload_slurm_queue(
    client_id: str,
    slurm_queue: list[dict[str, str]],
    respx_mock: MockRouter,
    mock_settings: mock.MagicMock,
    mock_logger: mock.MagicMock,
):
    """Test if the upload_slurm_queue send the proper GraphQL query to the Vantage API."""
    mock_settings.OIDC_CLIENT_ID = client_id
    mock_settings.BASE_API_URL = backend_client.base_url

    response_payload = {"data": {"uploadSlurmQueue": {"message": "Slurm queue uploaded successfully"}}}

    query = dedent(
        """
        mutation($queue: JSONScalar!, $clientId: String!) {
            uploadSlurmQueue(queue: $queue, clientId: $clientId) {
                ... on UploadSlurmQueueSuccess {
                    message
                }
            }
        }
        """
    )
    body = {"query": query, "variables": {"queue": slurm_queue, "clientId": client_id}}

    respx_mock.post("/cluster/graphql", json=body).respond(status_code=200, json=response_payload)

    await upload_slurm_queue(slurm_queue)

    mock_logger.debug.assert_any_call(f"Response from Vantage API while uploading queue: {response_payload}")



"""Core module for syncing Keycloak teams with SSSD."""

import re
import subprocess
from shutil import copyfile
from textwrap import dedent
from typing import Iterable

from vantage_agent.logger import logger
from vantage_agent.settings import SETTINGS
from vantage_agent.vantage_api_client import AsyncBackendClient


def create_backup(file_path: str):
    """Create a backup of the given file."""
    bkp_name = str(file_path).split("/").pop()
    backup_path = f"{SETTINGS.CACHE_DIR}/{bkp_name}.bkp"
    copyfile(file_path, backup_path)
    logger.info(f"Backup of {bkp_name} created at {backup_path}")


def _extract_org_id_from_filter(filter_value: str) -> str | None:
    """Extract the org_id from an ldap_access_filter value."""
    match = re.search(r"ou=([^,]+),ou=organizations", filter_value)
    if match:
        return match.group(1)
    return None


def _build_ldap_filter(org_id: str, team_names: Iterable[str]) -> str:
    """Build the ldap_access_filter value, always keeping slurm-users present."""
    slurm_group_filter = (
        f"(memberOf=cn=slurm-users,ou=Groups,ou={org_id},ou=organizations,dc=vantagecompute,dc=ai)"
    )
    team_filters = [
        f"(memberOf=cn={team},ou=Teams,ou={org_id},ou=organizations,dc=vantagecompute,dc=ai)"
        for team in sorted({name for name in team_names if name})
    ]
    return f"(|{slurm_group_filter}{''.join(team_filters)})"


def _resolve_org_id(teams: list[dict], sssd_lines: list[str]) -> str | None:
    """Pick the organization id from teams response or the existing SSSD filter."""
    for team in teams:
        org_id = team.get("organization")
        if org_id:
            return org_id

    for line in sssd_lines:
        if line.strip().startswith("ldap_access_filter"):
            _, _, filter_value = line.partition("=")
            org_id = _extract_org_id_from_filter(filter_value)
            if org_id:
                return org_id
    return None


def update_sssd_access_filter(org_id: str, team_names: list[str], sssd_lines: list[str]) -> bool:
    """Update ldap_access_filter in the sssd.conf file. Returns True if the file was changed."""
    create_backup(SETTINGS.SSSD_CONF_PATH)

    new_filter_value = _build_ldap_filter(org_id, team_names)
    new_line = f"ldap_access_filter = {new_filter_value}\n"

    updated_lines = list(sssd_lines)
    for idx, line in enumerate(updated_lines):
        if line.strip().startswith("ldap_access_filter"):
            if line.strip() == new_line.strip():
                logger.info("ldap_access_filter already up to date. No changes made to sssd.conf.")
                return False
            updated_lines[idx] = new_line
            break
    else:
        updated_lines.append(new_line)

    with open(SETTINGS.SSSD_CONF_PATH, "w") as f:
        f.writelines(updated_lines)

    logger.info("Updated ldap_access_filter in %s", SETTINGS.SSSD_CONF_PATH)
    return True


async def get_cluster_teams() -> list[dict]:
    """Fetch teams associated with the current cluster."""
    query = dedent(
        """
        query($clusterId: String!) {
            teamsByCluster(clusterId: $clusterId) {
                id
                name
                organization
            }
        }
        """
    )
    body = {"query": query, "variables": {"clusterId": SETTINGS.OIDC_CLIENT_ID}}

    async with AsyncBackendClient() as backend_client:
        response = await backend_client.post("/teams/graphql", json=body)
        response.raise_for_status()
        payload = response.json()
        print(payload)

    data = payload.get("data") or {}
    # Support both camelCase and snake_case keys just in case.
    teams = data.get("teamsByCluster")
    return teams


def reload_sssd():
    """Reload the SSSD to get the new changes in the sssd.conf."""
    try:
        logger.debug("Reloading SSSD config.")
        result = subprocess.run(
            [
                "systemctl", "stop", "sssd", "&&",
                "sss_cache", "-E", "&&",
                "rm", "-rf", "/var/lib/sss/db/*", "&&",
                "rm", "-rf", "/var/lib/sss/mc/*", "&&",
                "systemctl", "restart", "sssd"
            ], check=True, capture_output=True, text=True
        )

        logger.info(f"SSSD config reloaded:\n{result.stdout}")
    except subprocess.CalledProcessError as e:
        logger.error(f"Error while reloading SSSD conf:\n{e.stderr}")
        raise e


async def sync_cluster_teams():
    """Sync Keycloak teams with SSSD ldap_access_filter."""
    try:
        logger.info("Syncing SSSD teams access filter.")
        teams = await get_cluster_teams()
        team_names = [team.get("name", "") for team in teams]

        with open(SETTINGS.SSSD_CONF_PATH, "r") as f:
            sssd_lines = f.readlines()

        org_id = _resolve_org_id(teams, sssd_lines)
        if not org_id:
            logger.error(
                "Unable to determine organization id from teams or existing sssd.conf. Aborting update."
            )
            return

        updated = update_sssd_access_filter(org_id=org_id, team_names=team_names, sssd_lines=sssd_lines)
        if updated:
            logger.info(f"SSSD ldap_access_filter synced with {len(set(team_names))} team(s).")
            logger.debug("Reloading SSSD config.")
            reload_sssd()

    except FileNotFoundError:
        logger.error("SSSD configuration file not found at %s", SETTINGS.SSSD_CONF_PATH)
    except Exception as exc:
        logger.error(f"Error while syncing teams to SSSD: {exc}")



"""Task definitions for the Vantage Agent."""

from apscheduler.job import Job
from apscheduler.schedulers.base import BaseScheduler

from vantage_agent.logicals.health import report_health_status
from vantage_agent.logicals.queue_actions import process_cluster_queue_actions
from vantage_agent.logicals.scontrol import (
    upload_scontrol_config,
    upload_scontrol_node,
    upload_scontrol_partition,
)
from vantage_agent.logicals.squeue import upload_squeue_queue
from vantage_agent.logicals.sync_partitions import sync_cluster_partitions
from vantage_agent.logicals.sync_teams import sync_cluster_teams
from vantage_agent.logicals.update import self_update_agent
from vantage_agent.settings import SETTINGS


def cluster_config_task(scheduler: BaseScheduler) -> Job:
    """Schedule a task to report the cluster config."""
    return scheduler.add_job(upload_scontrol_config, "interval", seconds=SETTINGS.TASK_JOBS_INTERVAL_SECONDS)


def cluster_partition_task(scheduler: BaseScheduler) -> Job:
    """Schedule a task to report the cluster config."""
    return scheduler.add_job(
        upload_scontrol_partition, "interval", seconds=SETTINGS.TASK_JOBS_INTERVAL_SECONDS
    )


def cluster_queue_actions_task(scheduler: BaseScheduler) -> Job:
    """Schedule a task to process cluster queue actions (job cancellation)."""
    return scheduler.add_job(
        process_cluster_queue_actions, "interval", seconds=SETTINGS.TASK_JOBS_INTERVAL_SECONDS
    )


def cluster_queue_task(scheduler: BaseScheduler) -> Job:
    """Schedule a task to report the cluster queue."""
    return scheduler.add_job(upload_squeue_queue, "interval", seconds=SETTINGS.TASK_JOBS_INTERVAL_SECONDS)


def sync_cluster_partitions_task(scheduler: BaseScheduler) -> Job | None:
    """Schedule a task to report the cluster config."""
    if SETTINGS.IS_CLOUD_CLUSTER is False:
        return None
    return scheduler.add_job(sync_cluster_partitions, "interval", seconds=SETTINGS.TASK_JOBS_INTERVAL_SECONDS)


def sync_cluster_teams_task(scheduler: BaseScheduler) -> Job | None:
    """Schedule a task to report the cluster config."""
    if SETTINGS.IS_CLOUD_CLUSTER is True:
        return None
    return scheduler.add_job(sync_cluster_teams, "interval", seconds=SETTINGS.TASK_JOBS_INTERVAL_SECONDS)

def cluster_node_task(scheduler: BaseScheduler) -> Job:
    """Schedule a task to report the cluster config."""
    return scheduler.add_job(upload_scontrol_node, "interval", seconds=SETTINGS.TASK_JOBS_INTERVAL_SECONDS)


def self_update_task(scheduler: BaseScheduler) -> Job:
    """Schedule a task to self update the agent every ``TASK_SELF_UPDATE_INTERVAL_SECONDS`` seconds."""
    if SETTINGS.TASK_SELF_UPDATE_INTERVAL_SECONDS is None:
        return None
    return scheduler.add_job(
        self_update_agent, "interval", seconds=SETTINGS.TASK_SELF_UPDATE_INTERVAL_SECONDS
    )


def status_report_task(scheduler: BaseScheduler) -> Job:
    """Schedule a task to report the status."""
    return scheduler.add_job(report_health_status, "interval", seconds=SETTINGS.TASK_JOBS_INTERVAL_SECONDS)



"""Core module for testing the vantage_agent/sentry.py module."""

import uuid
from unittest import mock

import pytest
from _pytest.monkeypatch import MonkeyPatch
from sentry_sdk.utils import BadDsn

from vantage_agent.sentry import init_sentry


@pytest.fixture
def mock_logger(monkeypatch: MonkeyPatch) -> mock.MagicMock:
    """Mock the logger instance."""
    mock_logger = mock.Mock()
    monkeypatch.setattr("vantage_agent.sentry.logger", mock_logger)
    return mock_logger


@pytest.fixture
def mock_sentry_sdk(monkeypatch: MonkeyPatch) -> mock.MagicMock:
    """Mock the sentry_sdk instance."""
    mock_sentry_sdk = mock.Mock()
    monkeypatch.setattr("vantage_agent.sentry.sentry_sdk", mock_sentry_sdk)
    return mock_sentry_sdk


@pytest.fixture
def mock_settings(monkeypatch: MonkeyPatch) -> mock.MagicMock:
    """Mock the settings instance."""
    mock_settings = mock.Mock()
    monkeypatch.setattr("vantage_agent.sentry.SETTINGS", mock_settings)
    return mock_settings


class TestInitSentry:
    """Test cases for the init_sentry function."""

    def test_init_sentry_bad_dsn(self, mock_logger: mock.MagicMock, mock_sentry_sdk: mock.MagicMock) -> None:
        """Test the init_sentry function when the DSN is invalid."""
        mock_sentry_sdk.init.side_effect = BadDsn("Invalid DSN")

        init_sentry()

        mock_sentry_sdk.init.assert_called_once()
        mock_logger.debug.assert_called_once_with("Sentry could not be enabled: Invalid DSN")

    def test_init_sentry_success(
        self, mock_logger: mock.MagicMock, mock_sentry_sdk: mock.MagicMock, mock_settings: mock.MagicMock
    ) -> None:
        """Test the init_sentry function when the DSN is valid."""
        mock_settings.SENTRY_DSN = str(uuid.uuid4())

        init_sentry()

        mock_sentry_sdk.init.assert_called_once()
        mock_logger.debug.assert_called_once_with("Enabled Sentry since a valid DSN key was provided.")

    def test_init_sentry_empty_dsn(
        self, mock_logger: mock.MagicMock, mock_sentry_sdk: mock.MagicMock, mock_settings: mock.MagicMock
    ) -> None:
        """Test the init_sentry function when the DSN is empty."""
        mock_settings.SENTRY_DSN = ""

        init_sentry()

        mock_sentry_sdk.init.assert_called_once()
        mock_logger.debug.assert_called_once_with("Enabled Sentry since a valid DSN key was provided.")



"""Core module for agent health related operations.

Copied and adjusted from https://github.com/omnivector-solutions/jobbergate/blob/main/jobbergate-agent/jobbergate_agent/jobbergate/report_health.py.
"""

from textwrap import dedent

from loguru import logger

from vantage_agent.exceptions import VantageApiError
from vantage_agent.logger import log_error
from vantage_agent.settings import SETTINGS
from vantage_agent.vantage_api_client import AsyncBackendClient


async def report_health_status() -> None:
    """Ping the API to report the agent's status."""
    logger.debug("Reporting status to the API")
    with VantageApiError.handle_errors("Failed to report agent status", do_except=log_error):
        query = dedent(
            """\
            mutation ($clientId: String!, $interval: Int!) {
                reportAgentHealth(clientId: $clientId, interval: $interval)
            }"""
        )
        body = {
            "query": query,
            "variables": {
                "clientId": SETTINGS.OIDC_CLIENT_ID,
                "interval": SETTINGS.TASK_JOBS_INTERVAL_SECONDS,
            },
        }
        async with AsyncBackendClient() as backend_client:
            response = await backend_client.post("cluster/graphql", json=body)
        response.raise_for_status()



"""Core module for testing the vantage_agent/main.py module."""

import asyncio
from unittest import mock

import pytest
from _pytest.monkeypatch import MonkeyPatch

from vantage_agent.main import main


@pytest.fixture
def mock_init_scheduler(monkeypatch: MonkeyPatch):
    """Mock the init_scheduler function."""
    mock_scheduler = mock.Mock()
    monkeypatch.setattr("vantage_agent.main.init_scheduler", mock_scheduler)
    return mock_scheduler


@pytest.fixture
def mock_new_event_loop(monkeypatch: MonkeyPatch):
    """Mock the new_event_loop function."""
    mock_loop = mock.Mock()
    monkeypatch.setattr(asyncio, "new_event_loop", lambda: mock_loop)
    monkeypatch.setattr(asyncio, "set_event_loop", mock.Mock())
    return mock_loop


@pytest.fixture
def mock_logger(monkeypatch: MonkeyPatch):
    """Mock the logger object."""
    mock_logger = mock.Mock()
    monkeypatch.setattr("vantage_agent.main.logger", mock_logger)
    return mock_logger


@pytest.fixture
def mock_init_sentry(monkeypatch: MonkeyPatch):
    """Mock the init_sentry function."""
    mock_init_sentry = mock.Mock()
    monkeypatch.setattr("vantage_agent.main.init_sentry", mock_init_sentry)
    return mock_init_sentry


class TestMainFunction:
    """Test cases for the main function."""

    def test_normal_execution(
        self,
        mock_init_scheduler: mock.MagicMock,
        mock_new_event_loop: mock.MagicMock,
        mock_logger: mock.MagicMock,
        mock_init_sentry: mock.MagicMock,
    ):
        """Test main function under normal execution (no exceptions)."""

        # Make call_soon execute the callback immediately to trigger init_scheduler
        def execute_callback(callback):
            callback()

        mock_new_event_loop.call_soon.side_effect = execute_callback
        mock_new_event_loop.run_forever.side_effect = lambda: None

        main()

        mock_logger.info.assert_called_once_with("Starting the Vantage Agent")
        mock_init_scheduler.assert_called_once()
        mock_new_event_loop.call_soon.assert_called_once()
        mock_new_event_loop.run_forever.assert_called_once()
        mock_new_event_loop.close.assert_called_once()
        mock_init_sentry.assert_called_once_with()
        assert not mock_init_scheduler.return_value.shut_down.called



"""Tests for syncing Keycloak teams with SSSD."""

import json
from subprocess import CalledProcessError
from textwrap import dedent
from unittest import mock

import pytest
from respx.router import MockRouter

from vantage_agent.logicals.sync_teams import (
    _build_ldap_filter,
    _extract_org_id_from_filter,
    _resolve_org_id,
    create_backup,
    get_cluster_teams,
    reload_sssd,
    sync_cluster_teams,
    update_sssd_access_filter,
)
from vantage_agent.settings import SETTINGS


@mock.patch("vantage_agent.logicals.sync_teams.copyfile")
def test_create_backup(copyfile_mock: mock.Mock):
    """Ensure a backup file is created alongside the original."""
    file_path = "/etc/sssd/sssd.conf"

    create_backup(file_path)

    copyfile_mock.assert_called_once_with(file_path, f"{SETTINGS.CACHE_DIR}/sssd.conf.bkp")


@pytest.mark.parametrize(
    "filter_value,expected_org_id",
    [
        (
            "(|(memberOf=cn=slurm-users,ou=Groups,ou=org-123,ou=organizations,dc=vantagecompute,dc=ai))",
            "org-123",
        ),
        ("(memberOf=cn=slurm-users,ou=Groups,dc=vantagecompute,dc=ai)", None),
    ],
)
def test_extract_org_id_from_filter(filter_value: str, expected_org_id: str | None):
    """Extract the org_id from the ldap_access_filter value."""
    assert _extract_org_id_from_filter(filter_value) == expected_org_id


def test_build_ldap_filter_adds_slurm_group_and_sorts_teams():
    """Always include slurm-users and sort unique team names."""
    org_id = "org-abc"
    team_names = ["team-two", "team-one", "team-two", ""]

    filter_value = _build_ldap_filter(org_id, team_names)

    assert filter_value == (
        "(|"
        "(memberOf=cn=slurm-users,ou=Groups,ou=org-abc,ou=organizations,dc=vantagecompute,dc=ai)"
        "(memberOf=cn=team-one,ou=Teams,ou=org-abc,ou=organizations,dc=vantagecompute,dc=ai)"
        "(memberOf=cn=team-two,ou=Teams,ou=org-abc,ou=organizations,dc=vantagecompute,dc=ai)"
        ")"
    )


def test_resolve_org_id_prefers_team_payload():
    """Return org_id from teams before falling back to sssd.conf."""
    teams = [{"organization": "team-org"}, {"organization": "other-org"}]
    sssd_lines = [
        "ldap_access_filter = (|(memberOf=cn=slurm-users,ou=Groups,ou=sssd-org,ou=organizations,dc=vantagecompute,dc=ai))"  # noqa
    ]

    assert _resolve_org_id(teams, sssd_lines) == "team-org"


def test_resolve_org_id_from_sssd_lines_when_missing_in_teams():
    """Extract org_id from the existing ldap_access_filter line."""
    teams = [{"organization": None}, {}]
    sssd_lines = [
        "ldap_access_filter = (|(memberOf=cn=slurm-users,ou=Groups,ou=sssd-org,ou=organizations,dc=vantagecompute,dc=ai))\n"  # noqa
    ]

    assert _resolve_org_id(teams, sssd_lines) == "sssd-org"


def test_resolve_org_id_returns_none_when_not_found():
    """Return None when neither the payload nor sssd.conf contains org_id."""
    assert _resolve_org_id([], ["# no ldap_access_filter here\n"]) is None

@mock.patch("vantage_agent.logicals.sync_teams.create_backup")
@mock.patch("vantage_agent.logicals.sync_teams.SETTINGS")
def test_update_sssd_access_filter_updates_existing_line(
    settings_mock: mock.Mock,
    create_backup_mock: mock.Mock,
    tmp_path
):
    """Replace an existing ldap_access_filter entry."""
    sssd_conf_path = tmp_path / "sssd.conf"
    settings_mock.SSSD_CONF_PATH = sssd_conf_path
    settings_mock.CACHE_DIR = tmp_path
    sssd_lines = [
        "[sssd]\n",
        "ldap_access_filter = (|(memberOf=cn=slurm-users,ou=Groups,ou=old,ou=organizations,dc=vantagecompute,dc=ai))\n",  # noqa
    ]

    updated = update_sssd_access_filter("org-1", ["team-b", "team-a"], sssd_lines)

    expected_line = f"ldap_access_filter = {_build_ldap_filter('org-1', ['team-b', 'team-a'])}\n"
    with open(sssd_conf_path, "r") as f:
        lines = f.readlines()

    assert updated is True
    assert lines == ["[sssd]\n", expected_line]
    create_backup_mock.assert_called_once_with(settings_mock.SSSD_CONF_PATH)


@mock.patch("vantage_agent.logicals.sync_teams.create_backup")
@mock.patch("vantage_agent.logicals.sync_teams.SETTINGS")
def test_update_sssd_access_filter_appends_when_missing(
    settings_mock: mock.Mock, create_backup_mock: mock.Mock, tmp_path
):
    """Append ldap_access_filter when it is missing."""
    sssd_conf_path = tmp_path / "sssd.conf"
    settings_mock.SSSD_CONF_PATH = sssd_conf_path
    settings_mock.CACHE_DIR = tmp_path
    sssd_lines = ["[sssd]\n"]

    updated = update_sssd_access_filter("org-2", ["team-a"], sssd_lines)

    expected_line = f"ldap_access_filter = {_build_ldap_filter('org-2', ['team-a'])}\n"
    with open(sssd_conf_path, "r") as f:
        lines = f.readlines()

    assert updated is True
    assert lines == ["[sssd]\n", expected_line]
    create_backup_mock.assert_called_once_with(settings_mock.SSSD_CONF_PATH)


@mock.patch("vantage_agent.logicals.sync_teams.create_backup")
@mock.patch("vantage_agent.logicals.sync_teams.SETTINGS")
def test_update_sssd_access_filter_returns_false_when_no_change(
    settings_mock: mock.Mock, create_backup_mock: mock.Mock, tmp_path
):
    """Do not rewrite the file if the filter is already up to date."""
    sssd_conf_path = tmp_path / "sssd.conf"
    settings_mock.SSSD_CONF_PATH = sssd_conf_path
    settings_mock.CACHE_DIR = tmp_path
    existing_line = f"ldap_access_filter = {_build_ldap_filter('org-3', ['team-a'])}\n"

    sssd_conf_path.write_text(existing_line)
    sssd_lines = [existing_line]

    updated = update_sssd_access_filter("org-3", ["team-a"], sssd_lines)

    assert updated is False
    assert sssd_conf_path.read_text() == existing_line
    create_backup_mock.assert_called_once_with(settings_mock.SSSD_CONF_PATH)


@pytest.mark.respx(base_url=str(SETTINGS.BASE_API_URL))
@pytest.mark.usefixtures("mock_access_token")
async def test_get_cluster_teams(respx_mock: MockRouter):
    """Fetch teams for the current cluster via GraphQL."""
    query = dedent(
        """
        query($clusterId: String!) {
            teamsByCluster(clusterId: $clusterId) {
                id
                name
                organization
            }
        }
        """
    )
    body = {"query": query, "variables": {"clusterId": SETTINGS.OIDC_CLIENT_ID}}
    teams = [
        {"id": "1", "name": "alpha", "organization": "org-1"},
        {"id": "2", "name": "beta", "organization": "org-1"},
    ]
    respx_mock.post("/teams/graphql", json=body).respond(
        status_code=200, content=json.dumps({"data": {"teamsByCluster": teams}})
    )

    response = await get_cluster_teams()

    assert response == teams


@mock.patch("vantage_agent.logicals.sync_teams.update_sssd_access_filter")
@mock.patch("vantage_agent.logicals.sync_teams.get_cluster_teams")
@mock.patch("vantage_agent.logicals.sync_teams.SETTINGS")
async def test_sync_cluster_teams_updates_filter(
    settings_mock: mock.Mock,
    get_cluster_teams_mock: mock.Mock,
    update_sssd_access_filter_mock: mock.Mock,
    tmp_path,
):
    """Happy path: read sssd.conf, resolve org_id, and update the filter."""
    sssd_conf_path = tmp_path / "sssd.conf"
    settings_mock.SSSD_CONF_PATH = sssd_conf_path
    sssd_filter_line = (
        "ldap_access_filter = (|(memberOf=cn=slurm-users,ou=Groups,ou=old,ou=organizations,dc=vantagecompute,dc=ai))\n"  # noqa
    )
    sssd_conf_path.write_text(sssd_filter_line)

    teams = [
        {"id": "1", "name": "alpha", "organization": "org-xyz"},
        {"id": "2", "name": "beta", "organization": "org-xyz"},
    ]
    get_cluster_teams_mock.return_value = teams

    await sync_cluster_teams()

    get_cluster_teams_mock.assert_called_once_with()
    update_sssd_access_filter_mock.assert_called_once_with(
        org_id="org-xyz",
        team_names=["alpha", "beta"],
        sssd_lines=[sssd_filter_line],
    )


@mock.patch("vantage_agent.logicals.sync_teams.logger")
@mock.patch("vantage_agent.logicals.sync_teams.get_cluster_teams")
@mock.patch("vantage_agent.logicals.sync_teams.SETTINGS")
async def test_sync_cluster_teams_logs_missing_sssd_file(
    settings_mock: mock.Mock, get_cluster_teams_mock: mock.Mock, logger_mock: mock.Mock, tmp_path
):
    """Log an error when sssd.conf cannot be found."""
    settings_mock.SSSD_CONF_PATH = tmp_path / "missing.conf"
    get_cluster_teams_mock.return_value = []

    await sync_cluster_teams()

    logger_mock.error.assert_called_once_with(
        "SSSD configuration file not found at %s", settings_mock.SSSD_CONF_PATH
    )


@mock.patch("vantage_agent.logicals.sync_teams.logger")
@mock.patch("vantage_agent.logicals.sync_teams.update_sssd_access_filter")
@mock.patch("vantage_agent.logicals.sync_teams.get_cluster_teams")
@mock.patch("vantage_agent.logicals.sync_teams.SETTINGS")
async def test_sync_cluster_teams_logs_unexpected_errors(
    settings_mock: mock.Mock,
    get_cluster_teams_mock: mock.Mock,
    update_sssd_access_filter_mock: mock.Mock,
    logger_mock: mock.Mock,
    tmp_path,
):
    """Capture unexpected exceptions while syncing teams."""
    settings_mock.SSSD_CONF_PATH = tmp_path / "sssd.conf"
    settings_mock.CACHE_DIR = tmp_path
    sssd_filter_line = (
        "ldap_access_filter = (|(memberOf=cn=slurm-users,ou=Groups,ou=existing,ou=organizations,dc=vantagecompute,dc=ai))\n"  # noqa
    )
    settings_mock.SSSD_CONF_PATH.write_text(sssd_filter_line)

    get_cluster_teams_mock.return_value = [{"id": "1", "name": "alpha", "organization": "org-1"}]
    update_sssd_access_filter_mock.side_effect = RuntimeError("unexpected boom")

    await sync_cluster_teams()

    update_sssd_access_filter_mock.assert_called_once()
    logger_mock.error.assert_called_once_with("Error while syncing teams to SSSD: unexpected boom")

@mock.patch("vantage_agent.logicals.sync_teams.subprocess.run")
def test_reload_sssd__test_the_success_execution(subprocess_run_mock: mock.Mock):
    """Reload the sssd to get the new changes in the sssd.conf."""
    reload_sssd()
    subprocess_run_mock.assert_called_once_with(
        [
                "systemctl", "stop", "sssd", "&&",
                "sss_cache", "-E", "&&",
                "rm", "-rf", "/var/lib/sss/db/*", "&&",
                "rm", "-rf", "/var/lib/sss/mc/*", "&&",
                "systemctl", "restart", "sssd"
            ], check=True, capture_output=True, text=True
    )

@mock.patch("vantage_agent.logicals.sync_teams.subprocess.run")
def test_reload_sssd__test_the_filed_execution(subprocess_run_mock: mock.Mock):
    """Reload the SSSD to get the new changes in the sssd.conf."""
    subprocess_run_mock.side_effect = CalledProcessError(
        stderr="Dummy Error", output="Log......", returncode=1, cmd=""
    )

    try:
        reload_sssd()
        assert False
    except Exception as e:
        subprocess_run_mock.assert_called_once_with(
            [
                "systemctl", "stop", "sssd", "&&",
                "sss_cache", "-E", "&&",
                "rm", "-rf", "/var/lib/sss/db/*", "&&",
                "rm", "-rf", "/var/lib/sss/mc/*", "&&",
                "systemctl", "restart", "sssd"
            ], check=True, capture_output=True, text=True
        )
        assert isinstance(e, CalledProcessError)


@mock.patch("vantage_agent.logicals.sync_teams.reload_sssd")
@mock.patch("vantage_agent.logicals.sync_teams.update_sssd_access_filter")
@mock.patch("vantage_agent.logicals.sync_teams.get_cluster_teams")
@mock.patch("vantage_agent.logicals.sync_teams.SETTINGS")
async def test_sync_cluster_teams_returns_when_org_id_missing(
    settings_mock: mock.Mock,
    get_cluster_teams_mock: mock.Mock,
    update_sssd_access_filter_mock: mock.Mock,
    reload_sssd_mock: mock.Mock,
    tmp_path,
):
    """Abort when org_id cannot be resolved."""
    sssd_conf_path = tmp_path / "sssd.conf"
    settings_mock.SSSD_CONF_PATH = sssd_conf_path
    sssd_conf_path.write_text("# empty file\n")

    get_cluster_teams_mock.return_value = [{"id": "1", "name": "lonely", "organization": None}]

    await sync_cluster_teams()

    get_cluster_teams_mock.assert_called_once_with()
    update_sssd_access_filter_mock.assert_not_called()
    reload_sssd_mock.assert_not_called()



"""Main module of the project for starting the agent."""

import asyncio

from loguru import logger

from vantage_agent.scheduler import init_scheduler, shut_down_scheduler
from vantage_agent.sentry import init_sentry


def main():
    """Start the agent by initiating the scheduler."""
    logger.info("Starting the Vantage Agent")
    init_sentry()

    # Create event loop for Python 3.14 compatibility
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    # Initialize scheduler after event loop is set but before it runs
    # APScheduler 3.11+ requires the event loop to be running when start() is called
    # So we use call_soon to schedule the initialization
    loop.call_soon(lambda: init_scheduler())

    try:
        loop.run_forever()
    except (KeyboardInterrupt, SystemExit):  # pragma: no cover
        logger.info("Shutting down the Vantage Agent")
        # Get the scheduler instance - it should be available via the module-level variable
        from vantage_agent.scheduler import scheduler

        shut_down_scheduler(scheduler)  # pragma: no cover
    finally:
        loop.close()


if __name__ == "__main__":
    main()  # pragma: no cover
